'''This document contains a number of sample trackers.'''

import os
import random
import glob
import itertools
import pandas
import numpy
import collections

from CGATReport.Tracker import Tracker, Status
from collections import OrderedDict as odict


def BarData():
    return dict([("bar1", 20), ("bar2", 10)])


class LabeledDataExample(Tracker):
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2", "track3")

    def __call__(self, track, slice=None):

        v = int(track[-1]) * int(slice[-1])
        if slice == "slice1":
            return odict((("column1", v),
                          ("column2", v * 2),))
        elif slice == "slice2":
            return odict((("column1", v),
                          ("column2", v * 2),
                          ("column3", v * 3),))


class LabeledDataLargeExample(Tracker):
    slices = ("slice1", "slice2")
    tracks = ["track%i" % x for x in range(0, 100)]

    def __call__(self, track, slice=None):
        if slice == "slice1":
            return odict((("column1", 10),
                          ("column2", 20),))
        elif slice == "slice2":
            return odict((("column1", 20),
                          ("column2", 10),
                          ("column3", 5),))



class LabeledDataWithErrorsExample(Tracker):
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2", "track3")

    def __call__(self, track, slice=None):
        if slice == "slice1":
            return odict((
                ("column1", odict((('data', 20),
                                   ('error', 1),
                                   ))),
                ("column2", odict((('data', 10),
                                   ('error', 2),
                                   ))),
            ))
        elif slice == "slice2":
            return odict((
                ("column1", odict((('data', 20),
                                   ('error', 3),
                                   ))),
                ("column2", odict((('data', 10),
                                   ('error', 4)))),
                ("column3", odict((('data', 30),
                                   ('error', 5)))),
            ))


class LabeledDataWithErrorsAndLabelsExample(Tracker):

    '''use of ordered dictionary important as first
    element is treated as data.
    '''
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2", "track3")

    def __call__(self, track, slice=None):
        if slice == "slice1":
            return odict((
                ("column1", odict((('data', 20),
                                   ('error', 5),
                                   ('label', '**')))),
                ("column2", odict((('data', 10),
                                   ('error', 2),
                                   ('label', '*'))))
            ))
        elif slice == "slice2":
            return odict((
                ("column1", odict((('data', 20),
                                   ('error', 5),
                                   ('label', '***')))),
                ("column2", odict((('data', 10),
                                   ('error', 1)))),
                ("column3", odict((('data', 30),
                                   ('error', 4)))),
            ))


class HierarchicalLabeledDataExample(Tracker):
    '''labelled data with more than three levels.'''
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2", "track3")

    def __call__(self, track, slice=None):

        nvalues = 4
        values = collections.defaultdict(dict)
        levels = 3
        for x in range(nvalues):
            values[x] = collections.defaultdict(dict)
            for y in range(nvalues):
                values[x][y] = collections.defaultdict(dict)

        if slice == "slice1":
            for a, b, c in itertools.permutations(
                    range(nvalues),
                    levels):
                values[a][b][c] = (a + 1) * (b + 1) * (c + 1)
        elif slice == "slice2":
            for a, b, c in itertools.permutations(
                    range(nvalues),
                    levels):
                values[a][b][c] = (a + 1) * (b + 1) * (c + 1) * 2

        return values


class SingleColumnDataExample(Tracker):

    slices = ("slice1", "slice2")
    tracks = ("track1", "track2", "track3")

    def __call__(self, track, slice=None):
        s = [random.randint(0, 20) for x in range(40)]
        random.shuffle(s)
        return odict((("data", s),))


class ArrayDataExample(Tracker):

    '''return two arrays of data.'''

    slices = ["slice%i" % x for x in range(0, 2)]
    tracks = ["track%i" % x for x in range(0, 3)]

    def __call__(self, track, slice=None):

        scale = (3 - int(track[-1])) * (2 - int(slice[-1]))

        data = odict((("x", range(0, 10)),
                      ("y", [x * scale for x in range(0, 10)])))

        return data


class SingleColumnDataLargeExample(Tracker):

    '''return a single column of data.'''
    slices = ("slice1", "slice2")
    tracks = ["track%i" % x for x in range(0, 20)]

    def __call__(self, track, slice=None):
        s = [random.randint(0, 20) for x in range(40)]
        random.shuffle(s)
        return odict((("data", s),))


class SingleColumnDataWithErrorExample(Tracker):

    '''return a single column of data.'''
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2", "track3")

    def __call__(self, track, slice=None):
        s = [random.randint(0, 20) for x in range(40)]
        e = [random.randint(0, 3) for x in range(40)]
        random.shuffle(s)
        return odict((("data", s),
                      ("error", e)))


class SingleColumnDataExampleWithoutSlices(Tracker):

    '''return a single column of data.'''
    tracks = ("track1", "track2", "track3")

    def __call__(self, track, slice=None):
        s = [random.randint(0, 20) for x in range(40)]
        random.shuffle(s)
        return odict((("data", s),))


class MultipleColumnDataExample(Tracker):

    '''multiple columns each with a column with data.'''
    mColumns = ["col1", "col2", "col3"]
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2")

    def __call__(self, track, slice=None):
        data = []
        if slice == "slice1":
            for x in range(len(self.mColumns) - 1):
                data.append([y + random.gauss(0, 0.2) for y in range(10)])
        elif slice == "slice2":
            for x in range(len(self.mColumns)):
                data.append([y + random.gauss(0, 0.5) for y in range(10)])
        return odict(zip(self.mColumns, data))


class MultipleColumnDataFullExample(Tracker):

    '''multiple columns each with a column with data.'''
    mColumns = ["col1", "col2", "col3"]
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2")

    def __call__(self, track, slice=None):
        data = []
        if slice == "slice1":
            for x in range(len(self.mColumns)):
                data.append([y + random.gauss(0, 0.2) for y in range(20)])
        elif slice == "slice2":
            for x in range(len(self.mColumns)):
                data.append([y + random.gauss(0, 0.5) for y in range(20)])
        return odict(zip(self.mColumns, data))


class DeepLevelNestedIndexExample(Tracker):
    '''multiple columns each with a column with data.'''

    def __call__(self):

        index = pandas.MultiIndex.from_product(
            [list(x) for x in ["ABC", "DEF", "HIJ", "KLM", "XYZ"]])
        df = pandas.DataFrame()
        df["x"] = [y + random.gauss(0, 0.2) for y in range(len(index))]
        df["y"] = [y + random.gauss(0, 0.2) for y in range(len(index))]
        df.index = index
        return df


class DeepLevelNamedNestedIndexExample(Tracker):
    '''multiple columns each with a column with data.'''

    def __call__(self):

        index = pandas.MultiIndex.from_product(
            [list(x) for x in ["ABC", "DEF", "HIJ", "KLM", "XYZ"]],
            names=["level%i" % x for x in range(5)])
        df = pandas.DataFrame()
        df["x"] = [y + random.gauss(0, 0.2) for y in range(len(index))]
        df["y"] = [y + random.gauss(0, 0.2) for y in range(len(index))]
        df.index = index
        
        return df


class ErrorInTracker1(Tracker):

    '''A tracker that creates an error - exception while collecting data.'''
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2")

    def __call__(self, track, slice=None):
        raise ValueError("testing: could not collect data")


class ErrorInTracker2(Tracker):

    '''A tracker that creates an error - problems while returning tracks.'''
    slices = ("slice1", "slice2")

    def getTracks(self):
        raise ValueError("testing: could not build tracks")

    def __call__(self, track, slice=None):
        return odict((("data", range(0, 10)),))


class ErrorInTracker3(Tracker):

    '''A tracker that creates an error - no data.'''
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2")

    def __call__(self, track, slice=None):
        return None


class EmptyTracker(Tracker):

    '''A tracker that creates a warning.'''
    slices = ("slice1", "slice2")
    tracks = []

    def __call__(self, track, slice=None):
        return None


class LabeledDataTest(Tracker):
    slices = ("slice1", "slice2")
    tracks = ("track1", "track2", "track3")

    def __call__(self, track, slice=None):
        if slice == "slice1":
            return odict((("column1", 10),
                          ("column2", 20),))
        elif slice == "slice2":
            return odict((("column1", 20),
                          ("column2", 10),
                          ("column3", 5),))


class StatusTracker(Status):
    tracks = ("track1", "track2", "track3")

    def testTest1(self, track):
        '''test1 passes'''
        return "PASS", 0.5

    def testTest2(self, track):
        '''test2 fails -
        A large test.'''
        return "FAIL", 2

    def testTest3(self, track):
        '''test3 gives a warning'''
        return "WARN", "a string"

    def testTest4(self, track):
        '''test4 is not available/applicable'''
        return "NA", None


class MatrixTracker(Tracker):

    '''returns matrices.'''

    tracks = 'example1', 'example2'

    def __call__(self, track):

        if track == 'example1':
            matrix = numpy.arange(0, 10)
        elif track == 'example2':
            matrix = numpy.arange(10, 0, -1)

        matrix.shape = (2, 5)

        r = {'rows': list(map(str, range(0, 2))),
             'columns': list(map(str, range(0, 5))),
             'matrix': matrix}

        return r


def getSingleValue():
    return 12

IMAGEDIR = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "..", "images")


class DataWithImagesExample(Tracker):
    tracks = ("all",)

    def __call__(self, track):

        images = glob.glob(os.path.join(IMAGEDIR, "*.png"))

        data = odict()
        data["numbers"] = range(len(images))
        data["images"] = [".. image:: %s" % x for x in images]

        return data


class VennTracker(Tracker):

    tracks = ('two-circle', 'three-circle')
    slices = ('data1', 'data2')

    def __call__(self, track, slice):

        if slice == 'data1':
            f = 10
        else:
            f = 1

        if track == 'two-circle':
            return {'01': 10, '10': 20 * f, '11': 5 * f,
                    'labels': ("SetA", "SetB")}

        elif track == 'three-circle':
            return {'001': 10, '010': 20 * f, '100': 5 * f,
                    '011': 10, '110': 20 * f, '101': 5 * f,
                    '111': 10,
                    'labels': ("SetA", "SetB", "SetC")}



'''unit testing code for SphinxReport
'''

import itertools
from copy import deepcopy
import unittest

from CGATReport.DataTree import asDataFrame
from collections import OrderedDict as odict


class TestLabeledValues(unittest.TestCase):
    '''test dataframe conversions.'''

    def setUp(self):

        # order is explicitely non-alphabetical
        # to make sure it is preserved.

        self.data = odict((
            ('rowC', odict(
                (('colC', 1),
                 ('colB', 2),
                 ('colA', 3))
            )),
            ('rowB', odict(
                (('colC', 4),
                 ('colB', 5),
                 ('colA', 6))
            )),
            ('rowA', odict(
                (('colC', 7),
                 ('colB', 8),
                 ('colA', 9))
            ))
        ))

    def testSimple(self):

        df = asDataFrame(self.data)
        self.assertEqual(list(df.columns),
                         ["colC", "colB", "colA"])
        self.assertEqual(list(df.index),
                         ["rowC", "rowB", "rowA"])

    def testEmptyRow(self):

        for row in self.data.keys():
            dd = deepcopy(self.data)
            dd[row] = odict()
            df = asDataFrame(dd)
            self.assertEqual(list(df.columns),
                             ["colC", "colB", "colA"])
            self.assertEqual(list(df.index),
                             ["rowC", "rowB", "rowA"])

    def testMissingValues(self):

        for row, col in itertools.product(
                ("rowA", "rowB", "rowC"),
                ("colA", "colB", "colC")):
            dd = deepcopy(self.data)
            del dd[row][col]
            df = asDataFrame(dd)

            if row == "rowC" and col == "colB":
                # first row misses column B, so column A is second
                self.assertEqual(list(df.columns),
                                 ["colC", "colA", "colB"])
            elif row == "rowC" and col == "colC":
                # first row misses column C, so column B is second
                self.assertEqual(list(df.columns),
                                 ["colB", "colA", "colC"])
                
            else:
                self.assertEqual(list(df.columns),
                                 ["colC", "colB", "colA"])

            self.assertEqual(list(df.index),
                             ["rowC", "rowB", "rowA"])


class TestArrayValues(unittest.TestCase):
    '''test dataframe conversions.'''

    def setUp(self):

        # order is explicitely non-alphabetical
        # to make sure it is preserved.

        self.data = odict((
            ('rowC', odict(
                (('colC', [1] * 5),
                 ('colB', [2] * 5),
                 ('colA', [3] * 5))
            )),
            ('rowB', odict(
                (('colC', [4] * 6),
                 ('colB', [5] * 6),
                 ('colA', [6] * 6))
            )),
            ('rowA', odict(
                (('colC', [7] * 7),
                 ('colB', [8] * 7),
                 ('colA', [9] * 7))
            ))
        ))

        self.ref = ["rowC"] * 5 + ["rowB"] * 6 + ["rowA"] * 7

    def testSimple(self):

        df = asDataFrame(self.data)
        self.assertEqual(list(df.columns),
                         ["colC", "colB", "colA"])
        self.assertEqual(list(df.index),
                         self.ref)
        
    def testEmptyRow(self):

        for row in self.data.keys():
            dd = deepcopy(self.data)
            dd[row] = odict()
            df = asDataFrame(dd)
            self.assertEqual(list(df.columns),
                             ["colC", "colB", "colA"])
            self.assertEqual(list(df.index),
                             [x for x in self.ref if x != row])

    def testMissingValues(self):

        for row, col in itertools.product(
                ("rowA", "rowB", "rowC"),
                ("colA", "colB", "colC")):
            dd = deepcopy(self.data)
            del dd[row][col]
            df = asDataFrame(dd)

            if row == "rowC" and col == "colB":
                # first row misses column B, so column A is second
                self.assertEqual(list(df.columns),
                                 ["colC", "colA", "colB"])
            elif row == "rowC" and col == "colC":
                # first row misses column C, so column B is second
                self.assertEqual(list(df.columns),
                                 ["colB", "colA", "colC"])
                
            else:
                self.assertEqual(list(df.columns),
                                 ["colC", "colB", "colA"])

            self.assertEqual(list(df.index),
                             self.ref)

if __name__ == "__main__":
    unittest.main()



########################################################################
# Import setuptools
# Use existing setuptools
try:
    from setuptools import setup, find_packages
except ImportError:
    # try to get via ez_setup
    # ez_setup did not work on all machines tested as
    # it uses curl with https protocol, which is not
    # enabled in ScientificLinux
    import ez_setup
    ez_setup.use_setuptools()
    from setuptools import setup, find_packages

import glob
import sys
import os
import re
import distutils.sysconfig
import stat
import subprocess

major, minor1, minor2, s, tmp = sys.version_info


#####################################################################
# Code to install dependencies from a repository
#####################################################################
# Modified from http://stackoverflow.com/a/9125399
#####################################################################
def which(program):
    """
    Detect whether or not a program is installed.
    Thanks to http://stackoverflow.com/a/377028/70191
    """
    def is_exe(fpath):
        return os.path.exists(fpath) and os.access(fpath, os.X_OK)

    fpath, _ = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ['PATH'].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None

REPO_REQUIREMENT = re.compile(
    r'^-e (?P<link>(?P<vcs>git|svn|hg|bzr).+#egg=(?P<package>.+)-(?P<version>\d(?:\.\d)*))$')
HTTPS_REQUIREMENT = re.compile(
    r'^-e (?P<link>.*).+#(?P<package>.+)-(?P<version>\d(?:\.\d)*)$')
install_requires = []
dependency_links = []

for requirement in (l.strip() for l in open('requires.txt')
                    if not l.startswith("#")):
    match = REPO_REQUIREMENT.match(requirement)
    if match:
        assert which(match.group('vcs')) is not None, \
            "VCS '%(vcs)s' must be installed in order " \
            "to install %(link)s" % match.groupdict()
        install_requires.append("%(package)s==%(version)s" % match.groupdict())
        dependency_links.append(match.group('link'))
        continue

    if requirement.startswith("https"):
        install_requires.append(requirement)
        continue

    match = HTTPS_REQUIREMENT.match(requirement)
    if match:
        install_requires.append(
            "%(package)s>=%(version)s" % match.groupdict())
        dependency_links.append(match.group('link'))
        continue

    install_requires.append(requirement)

if major == 2:
    install_requires.extend(['web.py>=0.37',
                             'xlwt>=0.7.4',
                             'matplotlib-venn>=0.5'])
elif major == 3:
    pass

if major == 2 and minor1 < 5 or major < 2:
    raise SystemExit("""CGATReport requires Python 2.5 or later.""")

classifiers = """
Development Status :: 4 - Beta
Intended Audience :: Science/Research
Intended Audience :: Developers
License :: OSI Approved
Programming Language :: Python
Topic :: Software Development
Topic :: Scientific/Engineering
Operating System :: Microsoft :: Windows
Operating System :: POSIX
Operating System :: Unix
Operating System :: MacOS
"""

# external dependencies
# R
# sqlite
# R - ggplot2
# R - RSqlite
# R - gplots (for r-heatmap)
# graphvis - for dependency graphs in documentation

setup(name='CGATReport',
      version='0.3.2',
      description='CGATReport : a report generator in python based on sphinx',
      author='Andreas Heger',
      author_email='andreas.heger@gmail.com',
      packages=find_packages(),
      package_dir={'CGATReport': 'CGATReport',
                   'CGATReportPlugins': 'CGATReportPlugins'},
      url="https://github.com/AndreasHeger/CGATReport/",
      package_data={'CGATReport': [
          './templates/*.*',
          './templates/Makefile',
          './templates/js/*',
          './images/*']},
      license="BSD",
      platforms=["any"],
      keywords="report generator sphinx matplotlib sql",
      long_description='CGATReport : a report generator in python based '
      'on sphinx',
      classifiers=filter(None, classifiers.split("\n")),
      install_requires=install_requires,
      zip_safe=False,
      include_package_data=True,
      test_suite="tests",
      # python 3 conversion, requires distribute
      # use_2to3 = True,
      entry_points={
          'console_scripts': [
              'cgatreport-build = CGATReport.build:main',
              'cgatreport-clean = CGATReport.clean:main',
              'cgatreport-test = CGATReport.test:main',
              'cgatreport-quickstart = CGATReport.quickstart:main',
              'cgatreport-get = CGATReport.get:main',
              'cgatreport-profile = CGATReport.profile:main',
              'cgatreport-serve = CGATReport.serve:main',
          ],
          'distutils.commands': [
              'build_sphinx = cgatreport.setup_command:BuildDoc',
          ],
          'CGATReport.plugins': [
              'matplotlib='
              'CGATReportPlugins.MatplotlibPlugin:MatplotlibPlugin',
              'rplot='
              'CGATReportPlugins.RPlotPlugin:RPlotPlugin',
              'html='
              'CGATReportPlugins.HTMLPlugin:HTMLPlugin',
              'rst='
              'CGATReportPlugins.RSTPlugin:RSTPlugin',
              'xls='
              'CGATReportPlugins.XLSPlugin:XLSPlugin',
              'bokeh='
              'CGATReportPlugins.BokehPlugin:BokehPlugin',
              'transform-stats=CGATReportPlugins.Transformer:TransformerStats',
              'transform-correlation='
              'CGATReportPlugins.Transformer:TransformerCorrelationPearson',
              'transform-pearson='
              'CGATReportPlugins.Transformer:TransformerCorrelationPearson',
              'transform-contingency='
              'CGATReportPlugins.Transformer:TransformerContingency',
              'transform-spearman='
              'CGATReportPlugins.Transformer:TransformerCorrelationSpearman', 
              'transform-test-mwu='
              'CGATReportPlugins.Transformer:TransformerMannWhitneyU', 
              'transform-aggregate='
              'CGATReportPlugins.Transformer:TransformerAggregate',
              'transform-histogram='
              'CGATReportPlugins.Transformer:TransformerHistogram',
              # 'transform-tolabels=CGATReportPlugins.Transformer:TransformerToLabels',
              'transform-filter='
              'CGATReportPlugins.Transformer:TransformerFilter',
              # 'transform-indicator=CGATReportPlugins.Transformer:TransformerIndicator',
              # 'transform-select=CGATReportPlugins.Transformer:TransformerSelect',
              # 'transform-swop=CGATReportPlugins.Transformer:TransformerSwop',
              # 'transform-group=CGATReportPlugins.Transformer:TransformerGroup',
              # 'transform-combinations=CGATReportPlugins.Transformer:TransformerCombinations',
              # 'transform-combine=CGATReportPlugins.Transformer:TransformerCombinations',
              # 'transform-tolist=CGATReportPlugins.Transformer:TransformerToList',
              # 'transform-toframe=CGATReportPlugins.Transformer:TransformerToDataFrame',
              'transform-pandas='
              'CGATReportPlugins.Transformer:TransformerPandas',
              'transform-melt='
              'CGATReportPlugins.Transformer:TransformerMelt',
              'transform-pivot='
              'CGATReportPlugins.Transformer:TransformerPivot',
              # 'transform-count='
              # 'CGATReportPlugins.Transformer:TransformerCount',
              'transform-hypergeometric='
              'CGATReportPlugins.TransformersGeneLists:TransformerHypergeometric',
              # 'transform-label-paths='
              # 'CGATReportPlugins.TransformersGeneLists:TransformerPathToLabel',
              'transform-venn=CGATReportPlugins.TransformersGeneLists:TransformerVenn',
              'transform-p-adjust='
              'CGATReportPlugins.TransformersGeneLists:TransformerMultiTest',
              'transform-odds-ratio='
              'CGATReportPlugins.TransformersGeneLists:TransformerOddsRatio',
              'render-user='
              'CGATReportPlugins.Renderer:User',
              'render-debug='
              'CGATReportPlugins.Renderer:Debug',
              'render-dataframe='
              'CGATReportPlugins.Renderer:DataFrame',
              'render-table='
              'CGATReportPlugins.Renderer:Table',
              'render-rst-table='
              'CGATReportPlugins.Renderer:RstTable',
              'render-xls-table='
              'CGATReportPlugins.Renderer:XlsTable',
              'render-html-table='
              'CGATReportPlugins.Renderer:HTMLTable',
              'render-glossary-table='
              'CGATReportPlugins.Renderer:GlossaryTable',
              'render-matrix='
              'CGATReportPlugins.Renderer:TableMatrix',
              'render-matrixNP='
              'CGATReportPlugins.Renderer:NumpyMatrix',
              'render-status='
              'CGATReportPlugins.Renderer:Status',
              'render-line-plot='
              'CGATReportPlugins.Plotter:LinePlot',
              # for backwards compatibility
              'render-density-plot='
              'CGATReportPlugins.Seaborn:KdePlot',
              'render-histogram-plot='
              'CGATReportPlugins.Plotter:HistogramPlot',
              # 'render-histogram-gradient-plot='
              # 'CGATReportPlugins.Plotter:HistogramGradientPlot',
              'render-pie-plot='
              'CGATReportPlugins.Plotter:PiePlot',
              'render-scatter-plot=CGATReportPlugins.Plotter:ScatterPlot',
              'render-scatter-rainbow-plot='
              'CGATReportPlugins.Plotter:ScatterPlotWithColor',
              'render-matrix-plot='
              'CGATReportPlugins.Plotter:TableMatrixPlot',
              'render-matrixNP-plot='
              'CGATReportPlugins.Plotter:NumpyMatrixPlot',
              'render-hinton-plot='
              'CGATReportPlugins.Plotter:HintonPlot',
              'render-gallery-plot='
              'CGATReportPlugins.Plotter:GalleryPlot',
              'render-slideshow-plot='
              'CGATReportPlugins.SlideShow:SlideShowPlot',
              'render-bar-plot='
              'CGATReportPlugins.Plotter:BarPlot',
              'render-stacked-bar-plot='
              'CGATReportPlugins.Plotter:StackedBarPlot',
              'render-interleaved-bar-plot='
              'CGATReportPlugins.Plotter:InterleavedBarPlot',
              'render-venn-plot='
              'CGATReportPlugins.Plotter:VennPlot',
              # ggplot
              'render-ggplot='
              'CGATReportPlugins.GGPlotter:GGPlot',
              # pandas plotting
              'render-pdplot='
              'CGATReportPlugins.PandasPlotter:PandasPlot',
              # seaborn plots
              'render-sb-box-plot='
              'CGATReportPlugins.Seaborn:BoxPlot',
              'render-sb-violin-plot='
              'CGATReportPlugins.Seaborn:ViolinPlot',
              'render-sb-kde-plot='
              'CGATReportPlugins.Seaborn:KdePlot',
              'render-sb-pair-plot='
              'CGATReportPlugins.Seaborn:PairPlot',
              'render-sb-dist-plot='
              'CGATReportPlugins.Seaborn:DistPlot',
              'render-sb-heatmap-plot='
              'CGATReportPlugins.Seaborn:HeatmapPlot',
              'render-sb-clustermap-plot='
              'CGATReportPlugins.Seaborn:ClustermapPlot',
              # R plots
              'render-r-line-plot='
              'CGATReportPlugins.RPlotter:LinePlot',
              'render-r-box-plot='
              'CGATReportPlugins.RPlotter:BoxPlot',
              'render-r-smooth-scatter-plot='
              'CGATReportPlugins.RPlotter:SmoothScatterPlot',
              'render-r-heatmap-plot='
              'CGATReportPlugins.RPlotter:HeatmapPlot',
              'render-r-ggplot='
              'CGATReportPlugins.RPlotter:GGPlot',
              # Bokeh plots
              'render-b-line-plot='
              'CGATReportPlugins.BokehPlotter:LinePlot',
              # backwards compatibility
              'render-box-plot='
              'CGATReportPlugins.Seaborn:BoxPlot',
              'render-violin-plot='
              'CGATReportPlugins.Seaborn:ViolinPlot',
          ]
      },)

# fix file permission for executables
# set to "group writeable"
# also updates the "sphinx" permissions
if "install" in sys.argv:
    print ("updating file permissions for scripts")
    file_glob = os.path.join(
        distutils.sysconfig.project_base,
        "cgatreport-*")
    for x in glob.glob(file_glob):
        try:
            os.chmod(x, os.stat(x).st_mode | stat.S_IWGRP)
        except OSError:
            pass

    # replace the hardcoded python with /bin/env python. This
    # allows using the install within a virtual environment.
    print ("setting python to /bin/env python")
    statement = 'perl -p -i -e "s/\/ifs\/apps\/apps\/python-2.7.1\/bin\/python2.7/\/bin\/env python/" ' + file_glob
    print statement
    subprocess.call(statement, shell=True)
        



import os
import re
import random

from CGATReport.ResultBlock import ResultBlock, ResultBlocks
from CGATReportPlugins.Renderer import Renderer
from CGATReport.DataTree import path2str
from CGATReport import Utils
from docutils.parsers.rst import directives

import PIL

# Any scripts need to be added to layout.html in the theme.
# This will ensure that the relative path is used if a page
# is within a subdirectory. As a downside, all js code will
# be imported on every page.

# Plain setup, simple horizontal/vertical slider
# Customization: autoplay, dragorientation, size
PLAIN_SETUP = """
<script>
   jQuery(document).ready(function ($) {
            var options = {
                $AutoPlay: %(autoplay)s,
                $DragOrientation: 1,
            };
            var jssor_slider1 = new $JssorSlider$(
                                "%(name)s_container",
                                options);
        });
</script>

<div id="%(name)s_container"
        style="position: relative; top: 0px; left: 0px;
        width: %(width)ipx; height: %(width)ipx;">
<div u="slides"
        style="cursor: move; position: absolute; left: 0px; top: 0px;
        width: %(width)ipx; height: %(width)ipx; overflow: hidden;">
"""

LIST_SETUP = """
<script>
    jQuery(document).ready(function ($) {
        var options = {
            $AutoPlay: %(autoplay)s,
            $DragOrientation: 2,
            $ThumbnailNavigatorOptions: {
                    $Class: $JssorThumbnailNavigator$,
                    $ChanceToShow: 2,
                    $Loop: 2,
                    $AutoCenter: 3,
                    $Lanes: %(thumbnail_lanes)i,
                    $SpacingX: 4,
                    $SpacingY: 4,
                    $DisplayPieces: 4,
                    $ParkingPosition: 0,
                    $Orientation: 2,
                }
            };
        var jssor_slider1 = new $JssorSlider$(
             "%(name)s_container",
             options);
        });
</script>

<div id="%(name)s_container"
        style="position: relative; top: 0px; left: 0px;
        width: %(width)ipx; height: %(height)ipx;">
   <div u="slides"
        style="cursor: move; position: absolute; left: 0px; top: 0px;
        width: %(width)ipx; height: %(height)ipx; overflow: hidden;">

"""

LIST_SKIN = """
        <!-- ThumbnailNavigator Skin Begin -->
        <div u="thumbnavigator" class="jssort11" style="position: absolute;
              width: %(thumbnail_width)ipx; height: %(height)ipx;
              left:%(width)ipx; top:0px;">
            <!-- Thumbnail Item Skin Begin -->
            <style>
                /* jssor slider thumbnail navigator skin 11 css */
                /*
                .jssort11 .p            (normal)
                .jssort11 .p:hover      (normal mouseover)
                .jssort11 .pav          (active)
                .jssort11 .pav:hover    (active mouseover)
                .jssort11 .pdn          (mousedown)
                */
                .jssort11
                {
                        font-family: Arial, Helvetica, sans-serif;
                }
                .jssort11 .i, .jssort11 .pav:hover .i
                {
                        position: absolute;
                        top:3px;
                        left:3px;
                        WIDTH: 60px;
                        HEIGHT: 30px;
                        border: white 1px dashed;
                }
                * html .jssort11 .i
                {
                        WIDTH /**/: 62px;
                        HEIGHT /**/: 32px;
                }
                .jssort11 .pav .i
                {
                        border: white 1px solid;
                }
                .jssort11 .t, .jssort11 .pav:hover .t
                {
                        position: absolute;
                        top: 3px;
                        left: 68px;
                        width:129px;
                        height: 32px;
                        line-height:32px;
                        text-align: center;
                        color:#fc9835;
                        font-size:13px;
                        font-weight:700;
                }
                .jssort11 .pav .t, .jssort11 .phv .t, .jssort11 .p:hover .t
                {
                        color:#fff;
                }
                .jssort11 .c, .jssort11 .pav:hover .c
                {
                        position: absolute;
                        top: 38px;
                        left: 3px;
                        width: %(lane_width)ipx;
                        height: 31px;
                        line-height:31px;
                        color:#fff;
                        font-size:11px;
                        font-weight:400;
                        overflow: hidden;
                }
                .jssort11 .pav .c, .jssort11 .phv .c, .jssort11 .p:hover .c
                {
                        color:#fc9835;
                }
                .jssort11 .t, .jssort11 .c
                {
                        transition: color 2s;
                    -moz-transition: color 2s;
                    -webkit-transition: color 2s;
                    -o-transition: color 2s;
                    -webkit-transition: color 2s;
                    -o-transition: color 2s;
                }
                .jssort11 .p:hover .t, .jssort11 .phv .t, .jssort11
                .pav:hover .t, .jssort11 .p:hover .c, .jssort11 .phv .c,
                .jssort11 .pav:hover .c
                {
                        transition: none;
                    -moz-transition: none;
                    -webkit-transition: none;
                    -o-transition: none;
                }
                .jssort11 .p
                {
                        background:#181818;
                }
                .jssort11 .pav, .jssort11 .pdn
                {
                        background:#462300;
                }
                .jssort11 .p:hover, .jssort11 .phv, .jssort11 .pav:hover
                {
                        background:#333;
                }
            </style>
            <div u="slides" style="cursor: move;">
                <div u="prototype" class="p" style="position: absolute;
                     width: %(lane_width)ipx; height: 69px; top: 0; left: 0;">
                    <div u="thumbnailtemplate" style="
                     width: 100%%; height: 100%%;
                     border: none;position:absolute; top: 0; left: 0;"></div>
                </div>
            </div>
            <!-- Thumbnail Item Skin End -->
        </div>
        <!-- ThumbnailNavigator Skin End -->
"""


class SlideShowPlot(Renderer):
    """Factory class for slide show plots.

    The ``style`` option determines the slide show
    type that is returned.
    """
    
    options = Renderer.options +\
        (("autoplay", directives.flag),
         ("style", directives.unchanged),
         ("thumbnail-width", directives.positive_int),
         ("thumbnail-display", directives.positive_int),
         ("thumbnail-lanes", directives.positive_int))

    def __new__(cls, *args, **kwargs):

        # return special instances depending on style argument
        style = kwargs.get("style", "plain")

        if style == "plain":
            return PlainSlideShow(*args, **kwargs)
        elif style == "caption":
            return CaptionSlideShow(*args, **kwargs)
        elif style == "thumbnail-navigator":
            return ThumbnailSlideShow(*args, **kwargs)
        else:
            raise ValueError("unknown slide show style '%s'" % style)


class PlainSlideShow(Renderer):

    prefix = PLAIN_SETUP
    skin = ""

    # 1. DONE: automatically set paths and resources - through theme
    # 2. DONE: Multiple sliders per page without interference
    #       issue with checking if an element is up-to-date.
    #       check is for of images. Need to be different if
    #       to add the same element.
    # 3. DONE: Make sure images are packed into report
    # 4. DONE: get dimensions from rst directive to set size (width, height)
    # 5. DONE: Refactor boiler plate code - use factory
    # 6. implement different slideshows
    #    1. implement captions
    #    2. implement thumbnails
    # 7. Implement a non-html view

    # JS option: true if slide show should autoplay
    autoplay = "false"

    group_level = "all"
    nlevels = 1

    def __init__(self, *args, **kwargs):
        Renderer.__init__(self, *args, **kwargs)

        if "autoplay" in kwargs:
            self.autoplay = "true"
        else:
            self.autoplay = "false"

        self.thumbnail_display = kwargs.get("thumbnail-display", 4)
        self.thumbnail_lanes = kwargs.get("thumbnail-lanes", 1)
        self.thumbnail_width = kwargs.get("thumbnail-width", 200)

    def import_image(self, filename):
        '''import image into report. The image is hard-linked.

        returns path to image for use in html.
        '''

        mangled_filename = re.sub("/", "_", filename)
        filename = os.path.abspath(filename)

        # directory in which to store images and thumbnails
        outdir = Utils.getOutputDirectory()
        image_filename = os.path.join(outdir,
                                      mangled_filename)

        # filenames to use in html - must take document hierarchy
        # into account
        rst2srcdir = os.path.join(
            os.path.relpath(self.src_dir, start=self.rst_dir),
            outdir)
        html_image_filename = os.path.join(rst2srcdir,
                                           mangled_filename)

        try:
            os.link(filename, image_filename)
        except OSError:
            # file exists
            pass

        return html_image_filename

    def import_thumbnail(self, filename, thumbnail_size):
        '''import image in *filename* as a thumbnail.

        thumbnail_size is a tuple of (height, width) of the thumbnail
        in pixels.

        '''

        mangled_filename = re.sub("/", "_", filename)

        # directory in which to store images and thumbnails
        outdir = Utils.getOutputDirectory()
        thumb_filename = os.path.join(outdir,
                                      "thumb-%s.png" % mangled_filename)

        image = PIL.Image.open(filename)
        image.thumbnail(thumbnail_size)
        image.save(thumb_filename)

        # filenames to use in html - must take document hierarchy
        # into account
        rst2srcdir = os.path.join(
            os.path.relpath(self.src_dir, start=self.rst_dir),
            outdir)
        html_thumb_filename = os.path.join(rst2srcdir,
                                           "thumb-%s.png" % mangled_filename)

        return html_thumb_filename

    def add_image(self, filename, title, description):

        html_image_filename = self.import_image(filename)

        return [
            """<div><img u="image" src="%(html_image_filename)s" /></div>
            """ % locals()]

    def get_slideshow_options(self):

        self.width = int(self.display_options.get("width", 300))
        self.height = int(self.display_options.get("height", 300))
        return {'name': "SlideShowPlot_%09i" % random.randint(0, 1000000000),
                'width': self.width,
                'height': self.height,
                'autoplay': self.autoplay}

    def render(self, dataframe, path):

        blocks = ResultBlocks()

        options = self.get_slideshow_options()
        lines = [self.prefix % options]

        for title, row in dataframe.iterrows():
            row = row[row.notnull()]
            values = row.tolist()
            headers = list(row.index)

            dataseries = dict(zip(headers, values))
            try:
                # return value is a series
                filename = dataseries['filename']
            except KeyError:
                self.warn(
                    "no 'filename' key in path %s" % (path2str(path)))
                return blocks

            try:
                # return value is a series
                name = dataseries['name']
            except KeyError:
                self.warn(
                    "no 'name' key in path %s" % (path2str(path)))
                return blocks

            description, title = os.path.split(name)

            lines.extend(self.add_image(filename, title, description))

        lines.append("""</div>""")

        lines.append(self.skin % options)

        lines.append("""</div>""")

        lines = "\n".join(lines).split("\n")
        lines = [".. htmlonly::\n"] +\
            ["   .. raw:: html\n"] +\
            ["      " + x for x in lines]

        lines = "\n".join(lines)

        blocks.append(ResultBlock(text=lines,
                                  title=path2str(path)))
        return blocks


class CaptionSlideShow(PlainSlideShow):
    """A slide show with captions underneath the images."""
    
    def add_image(self, filename, title, description):

        html_image_filename = self.import_image(filename)

        width = self.width
        top = self.height - 50

        return [
            """<div>
                  <img u="image" src="%(html_image_filename)s"/>
                  <div style="position: absolute;
                       top: %(top)ipx; left: 0px;
                       width: %(width)ipx; height: 50px;
                       text-align:center;line-height:50px;
                       opacity: 0.5; filter: alpha(opacity=50);
                       background:#fff;">
                       %(title)s %(description)s
                  </div>
               </div>
            """ % locals()]


class ThumbnailSlideShow(PlainSlideShow):
    '''A slide show with column of thumbnails on the side.'''

    prefix = LIST_SETUP
    skin = LIST_SKIN

    def get_slideshow_options(self):
        options = PlainSlideShow.get_slideshow_options(self)
        options.update({
            "thumbnail_lanes": self.thumbnail_lanes,
            "thumbnail_display": self.thumbnail_display,
            "thumbnail_width": self.thumbnail_width,
            "lane_width": (self.thumbnail_width / self.thumbnail_lanes) - 3})
        return options

    def add_image(self, filename, title, description):

        thumbnail_size = self.height / self.thumbnail_display

        html_image_filename = self.import_image(filename)
        html_thumb_filename = self.import_thumbnail(
            filename,
            (thumbnail_size,
             thumbnail_size))

        return [
            """
            <div>
              <img u="image" src="%(html_image_filename)s" />
                <div u="thumb">
                  <abbr title="%(title)s %(description)s">
                     <img class="i" src="%(html_thumb_filename)s" />
                  </abbr>
                  <div class="c">%(title)s</div>
                  <div class="t">%(description)s</div>
                </div>
            </div>
            """ % locals()]




from logging import warn, log, debug, info
import itertools
import numpy
import pandas

# used in evals for computing bins in histograms
from numpy import arange

from collections import OrderedDict as odict
from CGATReport.Component import *
from CGATReport import Stats, DataTree, Utils

from docutils.parsers.rst import directives

# for rpy2 for data frames
try:
    import rpy2
    from rpy2.robjects import r as R
except ImportError:
    R = None

# ignore numpy histogram warnings in versions 1.3
import warnings


class Transformer(Component):

    '''Base class for transformers.

    Implements the basic __call__ method that iterates over
    a:term:`data tree` and calls self.transform method on the
    appropriate levels in the hierarchy.

    The property nlevels determines the grouping within a
    Transformer. If it is set to 0, data will be grouped by all
    levels in the index. If nlevels is None, the dataframe will not be
    grouped. If nlevels it is a positive number, a certain number of
    levels will be ignored from grouping.

    If nlevels is a positive number, the first nlevel levels will not
    be grouped. If it is a negative numbers, the last nlevel levels
    will be ignored for groupning.

    '''

    capabilities = ['transform']

    nlevels = None

    # If true, prune index in dataframe to
    # same levels as input
    prune_dataframe = True

    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, data):

        if self.nlevels is None:
            # do not group
            return self.transform(data)

        dataframes, keys = [], []
        group_levels = Utils.getGroupLevels(
            data,
            modify_levels=self.nlevels,
        )

        for key, group in data.groupby(level=group_levels):
            debug('applying transformation on group %s' % str(key))
            df = self.transform(group)
            if df is not None:
                dataframes.append(df)
                keys.append(key)

        df = pandas.concat(dataframes, keys=keys)
        
        if self.prune_dataframe:
            # reset dataframe index - keep the same levels
            Utils.pruneDataFrameIndex(df, original=data)

        debug("transform: finished")

        return df


########################################################################
########################################################################
# Conversion transformers
########################################################################
########################################################################
# class TransformerToLabels(Transformer):

#     '''convert:term:`numerical arrays` to:term:`labeled data`.

#     By default, the items are labeled numerically. If `tf-labels`
#     is given it is used instead.

#     Example::

#        Input:                 Returns:
#        a/keys/['x','y','z']   a/x/1
#        a/values/[4,5,6]       a/y/2
#                               a/z/3

#     Or::
#        Input:                       Returns:
#        a/contigs/['chr1','chr2']    a/chr1/10
#        a/lengths/[10,20]            a/chr2/20

#     Note that the outcome is equivalent to a pivot.

#     '''
#     nlevels = 1

#     options = Transformer.options +\
#         (('tf-labels', directives.unchanged),)

#     def __init__(self, *args, **kwargs):
#         Transformer.__init__(self, *args, **kwargs)

#         self.labels = kwargs.get("tf-labels", None)

#     def transform(self, data, path):
#         debug("%s: called" % str(self))

#         if len(data) == 0:
#             return data

#         keys = list(data.keys())
#         # pairs of data, see use case 2
#         if len(keys) == 2:
#             return odict(zip(*data.values()))

#         if self.labels:
#             labels = data[self.labels]
#             del keys[keys.index(self.labels)]
#             if len(keys) < 1:
#                 raise ValueError("TransformerToLabels requires at "
#                                  "least two arrays, got only 1, "
#                                  "if tf-labels is set")
#         else:
#             max_nkeys = max([len(x) for x in list(data.values())])
#             labels = list(range(1, max_nkeys + 1))

#         labels = list(map(str, labels))

#         if len(data) == 2:
#             new_data = odict(list(zip(labels, data[keys[0]])))
#         else:
#             new_data = odict()
#             for key in keys:
#                 new_data[key] = odict(list(zip(labels, data[key])))

#         return new_data


# class TransformerToList(Transformer):

#     '''transform categorized data into lists.

#     Example::

#        Input:            Returns:
#        a/x/1             x/[1,4]
#        a/y/2             y/[2,5]
#        a/z/3             z/[3,6]
#        b/x/4
#        b/y/5
#        b/z/6

#     '''
#     nlevels = 2

#     def __init__(self, *args, **kwargs):
#         Transformer.__init__(self, *args, **kwargs)

#     def transform(self, data, path):
#         debug("%s: called" % str(self))

#         lists = odict()

#         for major_key, values in data.items():
#             for minor_key, value in values.items():
#                 if minor_key in lists:
#                     lists[minor_key].append(value)
#                 else:
#                     lists[minor_key] = [value]

#         sizes = [len(x) for x in list(lists.values())]
#         if max(sizes) != min(sizes):
#             warn("%s: list of unequal sizes: min=%i, max=%i" %
#                  (self, min(sizes), max(sizes)))
#         return lists

# ########################################################################
# ########################################################################
# ########################################################################


# class TransformerToRDataFrame(Transformer):

#     '''transform data into one or more data frames.

#     Example::

#        Input:                                Output:
#        experiment1/expression = [1,2,3]      experiment1/df({ expression: [1,2,3], counts: [3,4,5] })
#        experiment1/counts = [3,4,5]          experiment2/df({ expression: [8,9,1], counts: [4,5,6] })
#        experiment2/expression = [8,9,1]
#        experiment2/counts = [4,5,6]

#     '''
#     nlevels = 1

#     def __init__(self, *args, **kwargs):
#         Transformer.__init__(self, *args, **kwargs)

#     def transform(self, data, path):
#         debug("%s: called" % str(self))

#         t = odict()
#         for minor_key, values in data.items():
#             if not Utils.isArray(values):
#                 raise ValueError("expected a list for data frame "
#                                  "creation, got %s", type(data))
#             if len(values) == 0:
#                 raise ValueError("empty list for %s" % (minor_key))
#             v = values[0]
#             if Utils.isInt(v):
#                 t[minor_key] = rpy2.robjects.IntVector(values)
#             elif Utils.isFloat(v):
#                 t[minor_key] = rpy2.robjects.FloatVector(values)
#             else:
#                 t[minor_key] = rpy2.robjects.StrVector(values)
 
#         return rpy2.robjects.DataFrame(t)


# class TransformerToDataFrame(Transformer):

#     '''transform data into one or more data frames.

#     Example::

#        Input:                                Output:
#        experiment1/expression = [1,2,3]      experiment1/df({ expression: [1,2,3], counts: [3,4,5] })
#        experiment1/counts = [3,4,5]          experiment2/df({ expression: [8,9,1], counts: [4,5,6] })
#        experiment2/expression = [8,9,1]
#        experiment2/counts = [4,5,6]

#     '''

#     def __init__(self, *args, **kwargs):
#         Transformer.__init__(self, *args, **kwargs)

#     def __call__(self, data):

#         result = DataTree.asDataFrame(data)
#         return odict((('all', result),))


# class TransformerIndicator(Transformer):

#     '''take a field from the lowest level and
#     build an absent/present indicator out of it.
#     '''

#     nlevels = 1
#     default = 0

#     options = Transformer.options +\
#         (('tf-fields', directives.unchanged),
#          ('tf-level', directives.length_or_unitless))

#     def __init__(self, *args, **kwargs):
#         Transformer.__init__(self, *args, **kwargs)

#         raise NotImplementedError('transformer indicator is not implemented')
#         try:
#             self.filter = kwargs["tf-fields"]
#         except KeyError:
#             raise KeyError("TransformerIndicator requires the "
#                            "`tf-fields` option to be set.")

#         try:
#             self.nlevels = int(kwargs["tf-level"])
#         except KeyError:
#             pass

#     def transform(self, data, path):
#         debug("%s: called" % str(self))

#         vals = data[self.filter]
#         return odict(list(zip(vals, [1] * len(vals))))


# class TransformerCount(Transformer):

#     '''compute counts of values in the hierarchy.

#     Displaying a table of counts can often be useful to
#     summarize the number of entries in a list prior to
#     plotting.

#     The following operations are perform when:term:`tf-level` is set
#     to ``1``::
#        Input:          Returns:
#        a/x=[1,2,3]            a/x=3
#        a/y=[1,3,5]            a/y=3
#        b/x=[34,3]             b/x=2
#        b/y=[2,4]              b/y=2
#     '''

#     nlevels = 1
#     default = 0

#     options = Transformer.options +\
#         (('tf-level', directives.length_or_unitless),)

#     def __init__(self, *args, **kwargs):
#         Transformer.__init__(self, *args, **kwargs)

#         try:
#             self.nlevels = int(kwargs["tf-level"])
#         except KeyError:
#             pass

#     def transform(self, data, path):
#         debug("%s: called" % str(self))

#         for v in list(data.keys()):
#             data[v] = len(data[v])

#         return data

########################################################################
########################################################################
# Filtering transformers
########################################################################
########################################################################
class TransformerPandas(Transformer):
    '''apply pandas dataframe methods to dataframe.

    '''

    nlevels = None
    default = 0

    options = Transformer.options +\
        (('tf-statement', directives.unchanged),)

    def __init__(self, *args, **kwargs):
        Transformer.__init__(self, *args, **kwargs)

        try:
            self.statement = kwargs["tf-statement"]
        except KeyError:
            raise KeyError("TransformerPandas requires the "
                           "`tf-statement` option to be set.")

    def transform(self, data):
        exec "data = data.{}".format(self.statement)
        return data


class TransformerFilter(Transformer):
    '''select columns from a dataframe.

    The columns are specified in the ``tf-fields`` option.

    '''

    nlevels = None
    default = 0

    options = Transformer.options +\
        (('tf-fields', directives.unchanged),)

    def __init__(self, *args, **kwargs):
        Transformer.__init__(self, *args, **kwargs)

        try:
            self.filter = kwargs["tf-fields"].split(",")
        except KeyError:
            raise KeyError("TransformerFilter requires the "
                           "`tf-fields` option to be set.")

    def transform(self, data):
        debug("%s: called" % str(self))
        return data[self.filter]


# class TransformerSelect(Transformer):

#     '''replace the lowest hierarchy with a single value.

#     This transformer removes all branches in a:term:`data tree` on
#     level:term:`tf-level` that do not match the :term:`tf-fields`
#     option.

#     The following operations are perform when:term:`tf-fields` is set
#     to ``x``::

#        Input:          Returns:
#        a/x=1            a=1
#        a/y=2            b=3
#        b/x=3
#        b/y=4

#     '''

#     nlevels = None
#     default = 0

#     options = Transformer.options +\
#         (('tf-fields', directives.unchanged),)

#     def __init__(self, *args, **kwargs):
#         Transformer.__init__(self, *args, **kwargs)

#         try:
#             self.fields = kwargs["tf-fields"].split(",")
#         except KeyError:
#             raise KeyError("TransformerSelect requires the "
#                            "`tf-fields` option to be set.")

#     def transform(self, data, path):
#         debug("%s: called" % str(self))

#         nfound = 0
#         for v in list(data.keys()):
#             for field in self.fields:
#                 try:
#                     data[v] = data[v][field]
#                     nfound += 1
#                     break
#                 except KeyError:
#                     pass
#             else:
#                 data[v] = self.default

#         if nfound == 0:
#             raise ValueError("could not find any field "
#                              "from `%s` in %s" %
#                              (str(self.fields), path))

#         return data


# class TransformerSwop(Transformer):

#     '''swop two levels in the data tree.

#     For example:

#     track1/gene_id=1, track1/gene_name=1
#     track2/gene_id=1, track2/gene_name=2

#     with tf-swop=0,1

#     will become:

#     '''
#     options = Transformer.options +\
#         (('tf-fields', directives.unchanged),)

#     def __init__(self, *args, **kwargs):
#         Transformer.__init__(self, *args, **kwargs)

#         try:
#             self.fields = kwargs["tf-fields"].split(",")
#         except KeyError:
#             raise KeyError("TransformerGroup requires the "
#                            "`tf-fields` option to be set.")

#         if len(self.fields) != 2:
#             raise ValueError("`tf-fields` requires exactly "
#                              "two fields for swapping")

#         self.fields = map(int, self.fields)

#     def __call__(self, data):
#         return DataTree.swop(data, self.fields[0], self.fields[1])


# class TransformerGroup(Transformer):

#     '''group second-to-last level by lowest level.

#     For example:

#     track1/gene_id=1, track1/gene_name=1
#     track2/gene_id=1, track2/gene_name=2

#     with tf-fields=gene_id

#     will become:

#     gene_id1/gene_name=1
#     gene_id1/tracks = [track1,track2]

#     Note that other fields not in the group field will take the value
#     of the first row.

#     For example::

#        Input:    Output:
#        a/x=1     x/y=1
#        a/y=1     x/tracks[a,b]
#        b/x=1
#        b/y=2

#     '''

#     nlevels = 2
#     default = 0

#     options = Transformer.options +\
#         (('tf-fields', directives.unchanged),)

#     def __init__(self, *args, **kwargs):
#         Transformer.__init__(self, *args, **kwargs)

#         try:
#             self.fields = kwargs["tf-fields"].split(",")
#         except KeyError:
#             raise KeyError("TransformerGroup requires the "
#                            "`tf-fields` option to be set.")

#         if len(self.fields) != 1:
#             raise ValueError("`tf-fields` requires exactly "
#                              "one field for grouping function")

#         self.field = self.fields[0]

#     def transform(self, data, path):
#         debug("%s: called" % str(self))

#         new_data = odict()

#         for v in list(data.keys()):
#             other_fields = [x for x in list(data[v].keys())
#                             if x != self.field]
#             for pos, val in enumerate(data[v][self.field]):
#                 if val not in new_data:
#                     new_data[val] = odict()
#                 if "group" not in new_data[val]:
#                     for o in other_fields:
#                         new_data[val][o] = data[v][o][pos]
#                     new_data[val]["group"] = ""
#                 new_data[val]["group"] += ",%s" % v

#         return new_data


# class TransformerCombinations(Transformer):

#     '''build combinations.

#     Level=2 can be used for labeled data::

#        Input:      Output:
#        a/x=1       a x b/a/x=1
#        b/x=2       a x b/b/x=2
#        c/x=3       a x c/a/x=1
#                    a x c/b/x=2
#                    b x c/a/x=2
#                    b x c/a/x=3

#     Uses the ``tf-fields`` option to combine a certain field.
#     Otherwise, it combines the first data found.

#     level=1 is useful to combine lists::

#        Input:            Output:
#        a/data=[1,2,3]    a x b/a=[1,2,3]
#        b/data=[2,4,5]    a x b/b=[2,4,5]
#        c/data=[4,2,1]    a x c/a=[1,2,3]
#                          a x c/a=[4,2,1]
#                          b x c/a=[2,4,5]
#                          b x c/a=[4,2,1]

#     '''

#     nlevels = 2

#     options = Transformer.options +\
#         (('tf-level', directives.length_or_unitless),
#          ('tf-fields', directives.unchanged),)

#     def __init__(self, *args, **kwargs):
#         Transformer.__init__(self, *args, **kwargs)

#         try:
#             self.fields = set(kwargs["tf-fields"].split(","))
#         except KeyError:
#             self.fields = None

#         self.nlevels = int(kwargs.get("tf-level", self.nlevels))

#     def transform(self, data, path):
#         debug("%s: called" % str(self))

#         vals = list(data.keys())
#         new_data = odict()

#         for x1 in range(len(vals) - 1):
#             n1 = vals[x1]
#             # find the first field that fits
#             if self.fields:
#                 for field in self.fields:
#                     if field in data[n1]:
#                         d1 = data[n1][field]
#                         break
#                 else:
#                     raise KeyError("could not find any match "
#                                    "from '%s' in '%s'" %
#                                    (str(list(data[n1].keys())),
#                                     str(self.fields)))
#             else:
#                 d1 = data[n1]

#             for x2 in range(x1 + 1, len(vals)):
#                 n2 = vals[x2]
#                 if self.fields:
#                     try:
#                         d2 = data[n2][field]
#                     except KeyErrror:
#                         raise KeyError("no field %s in '%s'" % sttr(data[n2]))
#                 else:
#                     d2 = data[n2]

#                 DataTree.setLeaf(new_data,
#                                  (("%s x %s" % (n1, n2)), n1),
#                                  d1)

#                 DataTree.setLeaf(new_data,
#                                  (("%s x %s" % (n1, n2)), n2),
#                                  d2)

#         return new_data


class TransformerStats(Transformer):
    '''Compute summary statistics for each
    column in a table.

    The summary statistics will be the columns
    of the data frame, while the column headers
    will be in the rows.
    '''

    # make sure that at least one grouping is done.
    nlevels = -1

    # keep row names (samples)
    prune_dataframe = False

    def __init__(self, *args, **kwargs):
        Transformer.__init__(self, *args, **kwargs)

    def transform(self, data):
        debug("%s: called" % str(self))
        return data.describe().transpose()


class TransformerPairwise(Transformer):

    '''for each pair of columns on the lowest level compute
    the pearson correlation coefficient and other stats.
    '''

    nlevels = 0
    method = None
    paired = False

    # This transformer increases the number of levels
    # so do not prune.
    prune_dataframe = False

    def __init__(self, *args, **kwargs):
        Transformer.__init__(self, *args, **kwargs)

    def transform(self, data):
        debug("%s: called" % str(self))

        if len(data.columns) < 2:
            raise ValueError("expected at least two columns, "
                             "got only %s." %
                             str(data.columns))

        results = []
        pairs = []

        for x, y in itertools.combinations(data.columns, 2):

            xvals, yvals = data[x], data[y]
            if self.paired and False:
                if len(xvals) != len(yvals):
                    raise ValueError("expected two arrays of the "
                                     "same length, %i != %i" %
                                     (len(xvals),
                                      len(yvals)))

                take = [i for i in range(len(xvals))
                        if xvals[i] is not None and yvals[i] is not None
                        and type(xvals[i]) in (float, int)
                        and type(yvals[i]) in (float, int)]
                xvals = [xvals[i] for i in take]
                yvals = [yvals[i] for i in take]

            try:
                result = self.apply(numpy.array(xvals),
                                    numpy.array(yvals))
            except ValueError as msg:
                warn("pairwise computation failed: %s" % msg)
                continue
            results.append(result)
            pairs.append((x, y))

        if results:
            df = pandas.DataFrame(results,
                                  index=pandas.MultiIndex.from_tuples(pairs))
        else:
            return None

        return df


class TransformerCorrelation(TransformerPairwise):

    '''compute correlation test

    Example::

       Input:
       set1=[1,2,3,4,5,6,7,8,9,10]
       set2=[3,4,5,6,7,8,9,10,11,12]
       set3=[5,6,7,8,9,10,11,12,14,15]

       Result:
       set1/set2/pvalue=0.0
       set1/set2/method=pearson
       set1/set2/nobservations=10
       set1/set2/coefficient=1.0
       set1/set2/alternative=two-sided
       set1/set2/logpvalue=0
       set1/set2/significance=***
    '''

    paired = True

    def apply(self, xvals, yvals):
        return Stats.doCorrelationTest(xvals, yvals,
                                       method=self.method)


class TransformerCorrelationPearson(TransformerCorrelation):

    '''for each pair of columns on the lowest level compute
    the pearson correlation coefficient and other stats.

    Example::

       Input:
       set1=[1,2,3,4,5,6,7,8,9,10]
       set2=[3,4,5,6,7,8,9,10,11,12]
       set3=[5,6,7,8,9,10,11,12,14,15]

       Output:
       set1/set2/pvalue=0.0
       set1/set2/method=pearson
       set1/set2/nobservations=10
       set1/set2/coefficient=1.0
       set1/set2/alternative=9
       set1/set2/logpvalue=0
       set1/set2/significance=***
       set1/set3/pvalue=
       ...
       set2/set3/pvalue=
       ...

    '''
    method = "pearson"


class TransformerCorrelationSpearman(TransformerCorrelation):

    '''for each pair of columns on the lowest level compute
    the spearman correlation coefficient and other stats.

    Example::

       Input:
       set1=[1,2,3,4,5,6,7,8,9,10]
       set2=[3,4,5,6,7,8,9,10,11,12]
       set3=[5,6,7,8,9,10,11,12,14,15]

       Output:
       set1/set2/pvalue=0.0
       set1/set2/method=spearman
       set1/set2/nobservations=10
       set1/set2/coefficient=1.0
       set1/set2/alternative=9
       set1/set2/logpvalue=0
       set1/set2/significance=***
       set1/set3/pvalue=0.0
       ...
       set2/set1/pvalue=0.0
       ...
       set2/set3/pvalue=0.0
       ...
       set3/set1/pvalue=0.0
       ...
       set3/set2/pvalue=0.0

    '''
    method = "spearman"


class TransformerMannWhitneyU(TransformerPairwise):

    '''apply the Mann-Whitney U test to test for
    the difference of medians.

    Example::

       Input:
       set1=[1,2,3,4,5,6,7,8,9,10]
       set2=[3,4,5,6,7,8,9,10,11,12]
       set3=[5,6,7,8,9,10,11,12,14,15]

       Output:


    '''

    def apply(self, xvals, yvals):
        xx = xvals[~numpy.isnan(xvals)]
        yy = yvals[~numpy.isnan(yvals)]
        r = Stats.doMannWhitneyUTest(xx, yy)
        return r


class TransformerContingency(TransformerPairwise):

    '''return number of identical entries

    Example::

       Input:
       set1=[1,2,3,4,5,6,7,8,9,10]
       set2=[3,4,5,6,7,8,9,10,11,12]
       set3=[5,6,7,8,9,10,11,12,14,15]

       Output:
       set1/set2=8
       set1/set3=6
       set2/set1=8
       set2/set3=8
       set3/set1=6
       set3/set2=8
    '''

    paired = False

    def apply(self, xvals, yvals):
        return len(set(xvals).intersection(set(yvals)))


class TransformerAggregate(Transformer):
    '''aggregate histogram like data.

    Possible aggregation options are:

    normalized-max
       normalize by maximum

    normalized-total
       normalize by total

    cumulative
       compute cumulative values

    reverse-cumulative
       compute reverse cumulative values

    relevel-with-first:
        add value of first bin to all other bins
        and set first bin to 0.

    '''

    nlevels = 0

    options = Transformer.options +\
        (('tf-aggregate', directives.unchanged),)

    def __init__(self, *args, **kwargs):
        Transformer.__init__(self, *args, **kwargs)

        self.mConverters = []
        self.mFormat = "%i"
        self.mBinMarker = "left"

        self.mMapKeyword = {
            "normalized-max": self.normalize_max,
            "normalized-total": self.normalize_total,
            "cumulative": self.cumulate,
            "reverse-cumulative": self.reverse_cumulate,
            "relevel-with-first": self.relevel_with_first,
        }

        if "tf-aggregate" in kwargs:
            for x in kwargs["tf-aggregate"].split(","):
                try:
                    self.mConverters.append(self.mMapKeyword[x])
                except KeyError:
                    raise KeyError("unknown keyword `%s`" % x)

        if self.normalize_total in self.mConverters or \
           self.normalize_max in self.mConverters:
            self.mFormat = "%6.4f"

        self.mBins = kwargs.get("tf-bins", "100")
        self.mRange = kwargs.get("tf-range", None)

        f = []
        if self.normalize_total in self.mConverters:
            f.append("relative")
        else:
            f.append("absolute")
        if self.cumulate in self.mConverters:
            f.append("cumulative")
        if self.reverse_cumulate in self.mConverters:
            f.append("cumulative")
        f.append("frequency")

        self.mYLabel = " ".join(f)

    def normalize_max(self, data):
        """normalize a data vector by maximum.
        """
        if data is None or len(data) == 0:
            return data
        m = max(data)
        data = data.astype(numpy.float)
        # numpy does not throw at division by zero, but sets values to Inf
        return data / m

    def relevel_with_first(self, data):
        """re-level data - add value of first bin to all other bins
        and set first bin to 0.
        """
        if data is None or len(data) == 0:
            return data
        v = data[0]
        data += v
        data[0] -= v
        return data

    def normalize_total(self, data):
        """normalize a data vector by the total"""
        if data is None or len(data) == 0:
            return data
        try:
            m = sum(data)
        except TypeError:
            return data
        data = data.astype(numpy.float)
        # numpy does not throw at division by zero, but sets values to Inf
        return data / m

    def cumulate(self, data):
        return data.cumsum()

    def reverse_cumulate(self, data):
        return data[::-1].cumsum()[::-1]

    def transform(self, data):
        debug("%s: called" % str(self))

        if len(data.columns) < 2:
            raise ValueError(
                'expected at least two columns, only got %s' %
                str(data.columns))

        for column in data.columns[1:]:
            for converter in self.mConverters:
                data.loc[:, column] = converter(data[column])

        return data


class TransformerHistogram(TransformerAggregate):
    '''compute a histogram of columns in the data
    frame.

    Returns a new dataframe with an additional
    first column called 'bins'.

    The bin-marker is placed on the left of the bin.
    '''

    nlevels = 0

    options = Transformer.options +\
        (('tf-bins', directives.unchanged),
         ('tf-range', directives.unchanged),
         ('tf-max-bins', directives.unchanged),
         )

    def __init__(self, *args, **kwargs):
        TransformerAggregate.__init__(self, *args, **kwargs)

        self.mFormat = "%i"
        self.mBinMarker = "left"

        self.mBins = kwargs.get("tf-bins", "100")
        self.mRange = kwargs.get("tf-range", None)
        self.max_bins = int(kwargs.get("max-bins", "1000"))

        f = []
        if self.normalize_total in self.mConverters:
            f.append("relative")
        else:
            f.append("absolute")
        if self.cumulate in self.mConverters:
            f.append("cumulative")
        if self.reverse_cumulate in self.mConverters:
            f.append("cumulative")
        f.append("frequency")

        self.mYLabel = " ".join(f)

    def binToX(self, bins):
        """convert bins to x-values."""
        if self.mBinMarker == "left":
            return bins[:-1]
        elif self.mBinMarker == "mean":
            return [(bins[x] - bins[x - 1]) / 2.0 for x in range(1, len(bins))]
        elif self.mBbinMarker == "right":
            return bins[1:]

    def toHistogram(self, data):
        '''compute the histogram.'''

        if len(data) == 0:
            warn("empty histogram")
            return None

        binsize = None

        if self.mRange is not None:
            vals = [x.strip() for x in self.mRange.split(",")]
            if len(vals) == 3:
                mi, ma, binsize = vals[0], vals[1], float(vals[2])
            elif len(vals) == 2:
                mi, ma, binsize = vals[0], vals[1], None
            elif len(vals) == 1:
                mi, ma, binsize = vals[0], None, None
            if mi is None or mi == "":
                mi = min(data.min())
            else:
                mi = float(mi)
            if ma is None or ma == "":
                ma = max(data.max())
            else:
                ma = float(ma)
        else:
            mi, ma = min(data.min()), max(data.max())

        if self.mBins.startswith("dict"):
            h = collections.defaultdict(int)
            for x in data:
                h[x] += 1
            bin_edges = sorted(h.keys())
            hist = numpy.zeros(len(bin_edges), numpy.int)
            for x in range(len(bin_edges)):
                hist[x] = h[bin_edges[x]]
            bin_edges.append(bin_edges[-1] + 1)
        else:
            if self.mBins.startswith("log"):

                try:
                    a, b = self.mBins.split("-")
                except ValueError:
                    raise SyntaxError("expected log-xxx, got %s" % self.mBins)
                nbins = float(b)
                if ma < 0 or mi < 0:
                    raise ValueError(
                        "can not bin logarithmically for negative values.")
                if mi == 0:
                    mi = numpy.MachAr().epsneg
                ma = numpy.log10(ma)
                mi = numpy.log10(mi)
                try:
                    bins = [10 ** x for x in numpy.arange(mi, ma, ma / nbins)]
                except ValueError as msg:
                    raise ValueError("can not compute %i bins for %f-%f: %s" %
                                     (nbins, mi, ma, msg))
            elif binsize is not None:
                # AH: why this sort statement? Removed
                # data.sort()

                # make sure that ma is part of bins
                bins = numpy.arange(mi, ma + binsize, binsize)
            else:
                try:
                    bins = eval(self.mBins)
                except SyntaxError as msg:
                    raise SyntaxError(
                        "could not evaluate bins from `%s`, error=`%s`"
                        % (self.mBins, msg))

            if hasattr(bins, "__iter__"):
                if len(bins) == 0:
                    warn("empty bins")
                    return None, None
                if self.max_bins > 0 and len(bins) > self.max_bins:
                    # truncate number of bins
                    warn("too many bins (%i) - truncated to (%i)" %
                         (len(bins), self.max_bins))
                    bins = self.max_bins

            # ignore histogram semantics warning
            all_counts = []
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                for column in data.columns:
                    counts, bin_edges = numpy.histogram(
                        data[column],
                        bins=bins, range=(mi, ma))
                    all_counts.append(counts)
            bin_edges = self.binToX(bin_edges)

            # re-build dataframe
            df = pandas.DataFrame.from_items(
                [('bin', bin_edges)] + zip(
                    data.columns, all_counts))

        return df

    def transform(self, data):
        debug("%s: called" % (str(self)))

        df = self.toHistogram(data)
        df = df.set_index("bin", append=True)
        for converter in self.mConverters:
            df = df.apply(converter, axis=0)

        df.reset_index(level="bin", inplace=True)
        debug("%s: completed" % (str(self)))
        return df


class TransformerMelt(Transformer):
    '''Create a melted table

    '''

    nlevels = None

    def __call__(self, data):
        ''' returns a melted table'''
        # merge index into dataframe
        # melt

        return pandas.melt(data.reset_index(), data.index.names)


class TransformerPivot(Transformer):
    '''pivot a table.

    Requires to define three columns in the
    dataframe to determine the row labels,
    column labels and values in the new dataframe.
    '''

    nlevels = None

    options = Transformer.options +\
        (('pivot-index', directives.unchanged),
         ('pivot-column', directives.unchanged),
         ('pivot-value', directives.unchanged),
         )

    pivot_index = None
    pivot_column = None
    pivot_value = None

    def __init__(self, *args, **kwargs):
        Transformer.__init__(self, *args, **kwargs)


        # convert to string, otherwise it is
        # docutils.nodes.reprunicode which does not evaluate as a
        # scalar in numpy and causes df.pivot() to fail.
        if "pivot-index" not in kwargs:
            raise ValueError('pivot requires a column to use as index')
        else:
            self.pivot_index = str(kwargs["pivot-index"])

        if "pivot-column" not in kwargs:
            raise ValueError('pivot requires a column to use as column')
        else:
            self.pivot_column = str(kwargs["pivot-column"])

        if "pivot-value" not in kwargs:
            raise ValueError('pivot requires a value to use as value')
        else:
            self.pivot_value = str(kwargs["pivot-value"])


    def __call__(self, data):
        ''' returns a melted table'''
        # merge index into dataframe
        # pivot

        return pandas.pivot_table(
            data.reset_index(),
            index=self.pivot_index,
            columns=self.pivot_column,
            values=self.pivot_value)
